import asyncio
import threading
from typing import Any, List


# noinspection PyShadowingBuiltins
def map(fn: callable, *args, mode="threading", **kwargs) -> Any:

    assert mode in (
        "threading",
        "asyncio",
    ), "map mode must be one of threading or asyncio."

    if args:
        num_calls = len(args[0])
    else:
        for v in kwargs.values():
            if isinstance(v, list):
                num_calls = len(v)
                break
        else:
            raise Exception(
                "At least one of the args or kwargs must be a list, "
                "which is to be mapped across the threads",
            )

    if mode == "threading":

        def fn_w_indexing(rets: List[None], thread_idx: int, *a, **kw):
            ret = fn(*a, **kw)
            rets[thread_idx] = ret

        threads = list()
        returns = [None] * num_calls
        for i in range(num_calls):
            a = tuple(a[i] for a in args)
            kw = {k: v[i] if isinstance(v, list) else v for k, v in kwargs.items()}
            thread = threading.Thread(
                target=fn_w_indexing,
                args=(returns, i, *a),
                kwargs=kw,
            )
            thread.start()
            threads.append(thread)
        [thread.join() for thread in threads]
        return returns

    # noinspection PyShadowingNames

    fns = []
    for i in range(num_calls):
        a = (a[i] for a in args)
        kw = {k: v[i] for k, v in kwargs.items()}
        fns.append(fn(*a, **kw))

    async def main():
        return await asyncio.gather(*fns)

    return asyncio.run(main())



# Implement #
# ----------#

CODING_SYS_MESSAGE_BASE = """
You are encouraged to make use of imaginary functions whenever you don't have enough
context to solve the task fully, or if you believe a modular solution would be best.
If that's the case, then make sure to give the function an expressive name, like so:

companies = get_all_companies_from_crm()
large_companies = filter_companies_based_on_headcount(
    companies, headcount=100, greater_than=True
)

Please DO NOT implement any inner functions. Just assume they exist, and make calls
to them like the examples above.

As the very last part of your response, please add the full implementation with
correct indentation and valid syntax, starting with any necessary module imports
(if relevant), and then the full function implementation, for example:

import {some_module}
from {another_module} import {function}

def {name}{signature}:
    {implementation}
"""
INIT_CODING_SYS_MESSAGE = """
You should write a Python implementation for a function `{name}` with the
following signature, docstring and example inputs:

def {name}{signature}
    \"\"\"
{docstring}
    \"\"\"

{name}({args} {kwargs})

    """
UPDATING_CODING_SYS_MESSAGE = """
You should *update* the Python implementation (preserving the function name,
arguments, and general structure), and only make changes as requested by the user
in the chat. The following example inputs should be compatible with the implementation:

{name}({args} {kwargs})

"""

DOCSTRING_SYS_MESSAGE_HEAD = """
We need to implement a new function `{name}`, but before we implement it we first
need to  decide on exactly how it should behave.
"""

DOCSTRING_SYS_MESSAGE_FIRST_CONTEXT = """
To help us, we know that the function `{child_name}` is called inside the function
`{parent_name}`, which has the following implementation:

```python
{parent_implementation}
```

Specifically, the line where `{child_name}` is called is: `{calling_line}`.
"""

DOCSTRING_SYS_MESSAGE_EXTRA_CONTEXT = """
This function (`{child_name}`) is itself called inside another function `{parent_name}`,
which has the following implementation:

```python
{parent_implementation}
```

Specifically, the line where `{child_name}` is called is: `{calling_line}`.
"""

DOCSTRING_SYS_MESSAGE_TAIL = """
Given all of this context, your task is to provide a well informed proposal for the
docstring and argument specification (with typing) for the new function `{name}`,
with an empty implementation `pass`, in the following format:

```python
def {name}({arg1}: {type1}, {arg2}: {type2} = {default2}, ...):
    \"\"\"
    {A very thorough description for exactly what this function needs to do.}

    Args:
        {arg1}: {arg1 description}

        {arg2}: {arg2 description}

    Returns:
        {return description}
    \"\"\"
    pass
```

Please respond in the format as above, and write nothing else after your answer.
"""

# Suggest #
# --------#

SUGGEST_SYS_MESSAGE = """

Your task is to propose changes to one parameter in the experiment configuration
in order to try and {relation} the metric `{metric}`. You should pay attention to all of
the log data, and pay special attention to examples where `{metric}` has a {low|high}
value. Try to work out why this might be, based on the full context of the experiment
configuration. You should then choose the experiment configuration parameter you'd like
to change (if there is more than one), and suggest a sensible new value to try.

The experiment configuration parameters that can be changed are as follows:

{configs}

The full set of evaluation logs for different combinations of these experiment
configuration parameters are as follows:

{evals}

You should think through this process step by step, and explain which parameter you've
chosen and why you think this parameter is contributing to the poor performing examples.

You should then propose a new value for this parameter, and why you believe this new
value will help. The parameter can be of any type. It might be a `str` system prompt,
a piece of Python code, a single numeric value, or dictionary, or any other type.

At the very end of your response, please respond as follows, filling in the placeholders
{parameter_name} and {parameter_value}:

"parameter":
"{parameter_name}"

"value":
{parameter_value}

"""



from __future__ import annotations
import inspect
import json
from contextvars import ContextVar
from typing import Any, Dict, List, Optional, Union

import requests
import unify
from unify import BASE_URL
from ...utils.helpers import _validate_api_key, _get_and_maybe_create_project

# Helpers #
# --------#

# trace
current_global_active_log = ContextVar("current_global_active_log", default=None)

# Context
current_global_active_log_kwargs = ContextVar(
    "current_global_active_kwargs",
    default={},
)
current_logged_logs = ContextVar("current_logged_logs_ids", default={})
current_context_nest_level = ContextVar("current_context_nest_level", default=0)

# span
current_span = ContextVar("current_span", default={})
running_time = ContextVar("running_time", default=0.0)


def _enclose_str(v):
    return json.dumps(v) if isinstance(v, str) else v


def _handle_special_types(
    kwargs: Dict[str, Any],
) -> Dict[str, Any]:
    new_kwargs = dict()
    for k, v in kwargs.items():
        if isinstance(v, unify.Dataset):
            v.upload()
            new_kwargs[k] = v.name
        elif callable(v):
            new_kwargs[k] = inspect.getsource(v)
        else:
            new_kwargs[k] = v
    return new_kwargs


def handle_multiple_logs(fn: callable):

    def wrapped(
        *args,
        logs: Optional[Union[int, unify.Log, List[Union[int, unify.Log]]]] = None,
        **kwargs,
    ):
        if logs is None and "logs" in fn.__code__.co_varnames:
            idx = fn.__code__.co_varnames.index("logs")
            if len(args) > idx:
                logs = args[idx]
                for i, val in enumerate(args[0:idx]):
                    kwargs[fn.__code__.co_varnames[i]] = val
                for i, val in enumerate(args[idx + 1 :]):
                    kwargs[fn.__code__.co_varnames[idx + i + 1]] = val
                args = ()
        if logs is None:
            current_active_log: Optional[unify.Log] = current_global_active_log.get()
            if current_active_log is None:
                raise Exception(
                    "If logs is unspecified, then current_global_active_log must be.",
                )
            return fn(*args, logs=current_active_log.id, **kwargs)
        elif isinstance(logs, int):
            return fn(*args, logs=logs, **kwargs)
        elif isinstance(logs, unify.Log):
            return fn(*args, logs=logs.id, **kwargs)
        elif isinstance(logs, list):
            if isinstance(logs[0], int):
                return unify.map(lambda lg: fn(*args, logs=lg, **kwargs), logs)
            elif isinstance(logs[0], unify.Log):
                return unify.map(lambda lg: fn(*args, logs=lg.id, **kwargs), logs)
            else:
                raise Exception(
                    f"list must contain int or unify.Log types, but found first entry "
                    f"{logs[0]} of type {type(logs[0])}",
                )
        else:
            raise Exception(
                f"logs argument must be of type int, unify.Log, or list, but found "
                f"{logs} of type {type(logs)}",
            )

    return wrapped


def log(
    project: Optional[str] = None,
    skip_duplicates: bool = True,
    api_key: Optional[str] = None,
    **kwargs,
) -> unify.Log:
    """
    Creates one or more logs associated to a project. unify.Logs are LLM-call-level data
    that might depend on other variables. This method returns the id of the new
    stored log.

    Args:
        project: Name of the project the stored logs will be associated to.

        skip_duplicates: Whether to skip creating new log entries for identical log
        data. If True (default), then the same eval Python script can be repeatedly run
        without duplicating the logged data every time. If False, then repeat entries
        will be added with identical data, but unique timestamps.

        api_key: If specified, unify API key to be used. Defaults to the value in the
        `UNIFY_KEY` environment variable.

        kwargs: Dictionary containing one or more key:value pairs that will be logged
        into the platform.

    Returns:
        The unique id of newly created log entry.
    """
    api_key = _validate_api_key(api_key)
    headers = {
        "accept": "application/json",
        "Authorization": f"Bearer {api_key}",
    }
    kwargs = {**kwargs, **current_global_active_log_kwargs.get()}
    kwargs = _handle_special_types(kwargs)
    project = _get_and_maybe_create_project(project, api_key=api_key)
    if skip_duplicates:
        retrieved_logs = get_logs_by_value(project, **kwargs, api_key=api_key)
        if retrieved_logs:
            assert len(retrieved_logs) == 1, (
                f"When skip_duplicates == True, then it's expected that each log "
                f"entry is unique, but found {len(retrieved_logs)} entries with "
                f"config {kwargs}"
            )
            return retrieved_logs[0]
    body = {"project": project, "entries": kwargs}
    response = requests.post(BASE_URL + "/log", headers=headers, json=body)
    response.raise_for_status()
    created_log = unify.Log(response.json(), api_key=api_key, **kwargs)
    if current_context_nest_level.get() > 0:
        current_logged_logs.set(
            {
                **current_logged_logs.get(),
                created_log.id: list(current_global_active_log_kwargs.get().keys()),
            },
        )
    return created_log


@handle_multiple_logs
def add_log_entries(
    logs: Optional[Union[int, unify.Log, List[Union[int, unify.Log]]]] = None,
    api_key: Optional[str] = None,
    **kwargs,
) -> Dict[str, str]:
    """
    Add extra entries into an existing log.

    Args:
        logs: The log(s) to update with extra data. Looks for the current active log if
        no id is provided.

        api_key: If specified, unify API key to be used. Defaults to the value in the
        `UNIFY_KEY` environment variable.

        kwargs: The data to log into the console.

    Returns:
        A message indicating whether the log was successfully updated.
    """
    log_id = logs  # handle_multiple_logs decorator handles logs, returning a single id
    current_active_log: Optional[unify.Log] = current_global_active_log.get()
    if current_active_log is None and log_id is None:
        raise ValueError(
            "`id` must be set if no current log is active within the context.",
        )
    api_key = _validate_api_key(api_key)
    headers = {
        "accept": "application/json",
        "Authorization": f"Bearer {api_key}",
    }
    if current_context_nest_level.get() > 0:
        kwargs = {
            **kwargs,
            **{
                k: v
                for k, v in current_global_active_log_kwargs.get().items()
                if k not in current_logged_logs.get().get(log_id, {})
            },
        }
    kwargs = {**kwargs, **current_global_active_log_kwargs.get()}
    kwargs = _handle_special_types(kwargs)
    body = {"entries": kwargs}
    # ToDo: remove this once duplicates are prevented in the backend
    current_keys = get_log_by_id(
        log_id if log_id else current_active_log.id,
    ).entries.keys()
    assert not any(key in body["entries"] for key in current_keys), (
        "Duplicate keys detected, please use replace_log_entries or "
        "update_log_entries if you want to replace or modify an existing key."
    )
    # End ToDo
    response = requests.put(
        BASE_URL + f"/log/{log_id if log_id else current_active_log.id}",
        headers=headers,
        json=body,
    )
    response.raise_for_status()
    return response.json()


@handle_multiple_logs
def delete_log(
    logs: Optional[Union[int, unify.Log, List[Union[int, unify.Log]]]] = None,
    api_key: Optional[str] = None,
) -> Dict[str, str]:
    """
    Deletes logs from a project.

    Args:
        logs: log(s) to delete from a project.

        api_key: If specified, unify API key to be used. Defaults to the value in the
        `UNIFY_KEY` environment variable.

    Returns:
        A message indicating whether the logs were successfully deleted.
    """
    log_id = logs  # handle_multiple_logs decorator handles logs, returning a single id
    api_key = _validate_api_key(api_key)
    headers = {
        "accept": "application/json",
        "Authorization": f"Bearer {api_key}",
    }
    response = requests.delete(BASE_URL + f"/log/{log_id}", headers=headers)
    response.raise_for_status()
    return response.json()


@handle_multiple_logs
def delete_log_entry(
    entry: str,
    logs: Optional[Union[int, unify.Log, List[Union[int, unify.Log]]]] = None,
    api_key: Optional[str] = None,
) -> Dict[str, str]:
    """
    Deletes an entry from a log.

    Args:
        entry: Name of the entries to delete from a given log.

        logs: log(s) to delete entries from.

        api_key: If specified, unify API key to be used. Defaults to the value in the
        `UNIFY_KEY` environment variable.

    Returns:
        A message indicating whether the log entries were successfully deleted.
    """
    log_id = logs  # handle_multiple_logs decorator handles logs, returning a single id
    api_key = _validate_api_key(api_key)
    headers = {
        "accept": "application/json",
        "Authorization": f"Bearer {api_key}",
    }
    entry = entry.replace("/", "-")
    response = requests.delete(
        BASE_URL + f"/log/{log_id}/entry/{entry}",
        headers=headers,
    )
    response.raise_for_status()
    return response.json()


@handle_multiple_logs
def replace_log_entries(
    logs: Optional[Union[int, unify.Log, List[Union[int, unify.Log]]]] = None,
    api_key: Optional[str] = None,
    **kwargs,
) -> Dict[str, str]:
    """
    Replaces existing entries in an existing log.

    Args:
        logs: The log(s) to replace fields for. Looks for the current active log if none
        specified.

        api_key: If specified, unify API key to be used. Defaults to the value in the
        `UNIFY_KEY` environment variable.

        kwargs: The data to update in the log.

    Returns:
        A message indicating whether the log was successfully updated.
    """
    log_id = logs  # handle_multiple_logs decorator handles logs, returning a single id
    api_key = _validate_api_key(api_key)
    for k, v in kwargs.items():
        delete_log_entry(k, log_id, api_key=api_key)
    return add_log_entries(log_id, api_key=api_key, **kwargs)


@handle_multiple_logs
def update_log_entries(
    fn: Union[callable, Dict[str, callable]],
    logs: Optional[Union[int, unify.Log, List[Union[int, unify.Log]]]] = None,
    api_key: Optional[str] = None,
    **kwargs,
) -> Dict[str, str]:
    """
    Updates existing entries in an existing log.

    Args:
        fn: The function or set of functions to apply to each field in the log.

        logs: The log(s) to update fields for. Looks for the current active log if not
        provided.

        api_key: If specified, unify API key to be used. Defaults to the value in the
        `UNIFY_KEY` environment variable.

        kwargs: The data to update in the log.

    Returns:
        A message indicating whether the log was successfully updated.
    """
    log_id = logs  # handle_multiple_logs decorator handles logs, returning a single id
    data = get_log_by_id(log_id, api_key=api_key).entries
    replacements = dict()
    for k, v in kwargs.items():
        f = fn[k] if isinstance(fn, dict) else fn
        replacements[k] = f(data[k], v)
    return replace_log_entries(log_id, api_key=api_key, **replacements)


@handle_multiple_logs
def rename_log_entries(
    logs: Optional[Union[int, unify.Log, List[Union[int, unify.Log]]]] = None,
    api_key: Optional[str] = None,
    **kwargs,
) -> Dict[str, str]:
    """
    Renames the set of log entries.

    Args:
        logs: The log(s) to update the field names for. Looks for the current active log
        if none are provided.

        api_key: If specified, unify API key to be used. Defaults to the value in the
        `UNIFY_KEY` environment variable.

        kwargs: The field names to update in the log, with keys as old names and values
        as new names.

    Returns:
        A message indicating whether the log field names were successfully updated.
    """
    log_id = logs  # handle_multiple_logs decorator handles logs, returning a single id
    api_key = _validate_api_key(api_key)
    data = get_log_by_id(log_id, api_key=api_key).entries
    for old_name in kwargs.keys():
        delete_log_entry(old_name, log_id, api_key=api_key)
    new_entries = {new_name: data[old_name] for old_name, new_name in kwargs.items()}
    return add_log_entries(log_id, api_key=api_key, **new_entries)


def get_logs(
    project: Optional[str] = None,
    filter: Optional[str] = None,
    api_key: Optional[str] = None,
) -> List[unify.Log]:
    """
    Returns a list of filtered logs from a project.

    Args:
        project: Name of the project to get logs from.

        filter: Boolean string to filter logs, for example:
        "(temperature > 0.5 and (len(system_msg) < 100 or 'no' in usr_response))"

        api_key: If specified, unify API key to be used. Defaults to the value in the
        `UNIFY_KEY` environment variable.

    Returns:
        The list of logs for the project, after optionally applying filtering.
    """
    api_key = _validate_api_key(api_key)
    headers = {
        "accept": "application/json",
        "Authorization": f"Bearer {api_key}",
    }
    project = _get_and_maybe_create_project(project, api_key=api_key)
    params = {
        "project": project,
        "filter_expr": filter,
    }
    response = requests.get(BASE_URL + "/logs", headers=headers, params=params)
    response.raise_for_status()
    params, logs = response.json().values()
    return [
        unify.Log(
            id=dct["id"],
            timestamp=dct["ts"],
            **dct["entries"],
            parameters={k: params[k][v] for k, v in dct["params"].items()},
            api_key=api_key,
        )
        for dct in logs
    ]


def delete_logs(
    project: Optional[str] = None,
    filter: Optional[str] = None,
    api_key: Optional[str] = None,
):
    """
    Returns a list of filtered logs from a project.

    Args:
        project: Name of the project to delete logs from.

        filter: Boolean string to filter logs for deletion, for example:
        "(temperature > 0.5 and (len(system_msg) < 100 or 'no' in usr_response))"

        api_key: If specified, unify API key to be used. Defaults to the value in the
        `UNIFY_KEY` environment variable.

    Returns:
        The list of deleted logs for the project, after optionally applying filtering.
    """
    logs = get_logs(project, filter, api_key=api_key)
    for log in logs:
        log.delete()
    return logs


def get_logs_with_fields(
    *fields: str,
    mode: str = "all",
    project: Optional[str] = None,
    api_key: Optional[str] = None,
) -> List[unify.Log]:
    """
    Returns a list of logs which contain the specified fields, either the logs which
    contain all of them ("all") or the logs which contain any of the fields ("any").

    Args:
        fields: The fields to retrieve logs for.

        mode: The retrieval mode, either returning the logs with all of the fields or
        the logs with any of the fields.

        project: Name of the project to get logs from.

        api_key: If specified, unify API key to be used. Defaults to the value in the
        `UNIFY_KEY` environment variable.

    Returns:
        A list of logs which contain the specified fields.
    """
    api_key = _validate_api_key(api_key)
    mode = {"any": "or", "all": "and"}[mode]
    filter_exp = f" {mode} ".join([f"exists({field})" for field in fields])
    return get_logs(project, filter=filter_exp, api_key=api_key)


def get_logs_without_fields(
    *fields: str,
    mode: str = "all",
    project: Optional[str] = None,
    api_key: Optional[str] = None,
) -> List[unify.Log]:
    """
    Returns a list of logs which do not contain the specified fields, either the logs
    which do not contain all of the fields ("all") or the logs which do not contain any
    of the fields ("any").

    Args:
        fields: The fields to not retrieve logs for.

        mode: The retrieval mode, either returning the logs with do not contain all of
        the fields or the logs which do not contain any of the fields.

        project: Name of the project to get logs from.

        api_key: If specified, unify API key to be used. Defaults to the value in the
        `UNIFY_KEY` environment variable.

    Returns:
        A list of logs which do not contain the specified fields.
    """
    api_key = _validate_api_key(api_key)
    mode = {"any": "and", "all": "or"}[mode]
    filter_exp = f" {mode} ".join([f"(not exists({field}))" for field in fields])
    return get_logs(project, filter=filter_exp, api_key=api_key)


def get_log_by_id(id: int, api_key: Optional[str] = None) -> unify.Log:
    """
    Returns the log associated with a given id.

    Args:
        id: IDs of the logs to fetch.

        api_key: If specified, unify API key to be used. Defaults to the value in the
        `UNIFY_KEY` environment variable.

    Returns:
        The full set of log data.
    """
    api_key = _validate_api_key(api_key)
    headers = {
        "accept": "application/json",
        "Authorization": f"Bearer {api_key}",
    }
    response = requests.get(BASE_URL + f"/log/{id}", headers=headers)
    response.raise_for_status()
    return unify.Log(id, **response.json()["entries"])


def get_logs_by_value(
    project: str,
    api_key: Optional[str] = None,
    **kwargs,
) -> List[unify.Log]:
    """
    Returns the logs with the data matching exactly if it exists,
    otherwise returns None.

    Args:
        project: Name of the project to get logs from.

        api_key: If specified, unify API key to be used. Defaults to the value in the
        `UNIFY_KEY` environment variable.

        Whether or not to include logs which contain identical key-value pairs to all
        kwargs passed which are present in the log, but

        kwargs: The data to search the upstream logs for.

    Returns:
        The list of unify.Logs which match the data, if any exist.
    """
    filter_str = " and ".join(
        [f"({k} == {_enclose_str(v)})" for k, v in kwargs.items()],
    )
    return get_logs(project, filter_str, api_key=api_key)


def get_log_by_value(
    project: str,
    api_key: Optional[str] = None,
    **kwargs,
) -> Optional[unify.Log]:
    """
    Returns the log with the data matching exactly if it exists,
    otherwise returns None.

    Args:
        project: Name of the project to get logs from.

        api_key: If specified, unify API key to be used. Defaults to the value in the
        `UNIFY_KEY` environment variable.

        kwargs: The data to search the upstream logs for.

    Returns:
        The single unify.Log which matches the data, if it exists.
    """
    logs = get_logs_by_value(project, **kwargs, api_key=api_key)
    assert len(logs) in (
        0,
        1,
    ), "Expected exactly zero or one log, but found {len(logs)}"
    return logs[0] if logs else None


def get_groups(
    key: str,
    project: Optional[str] = None,
    api_key: Optional[str] = None,
) -> Dict[str, List[Dict[str, Any]]]:
    """
    Returns a list of the different version/values of one entry within a given project
    based on its key.

    Args:
        key: Name of the log entry to do equality matching for.

        project: Name of the project to get logs from.

        api_key: If specified, unify API key to be used. Defaults to the value in the
        `UNIFY_KEY` environment variable.

    Returns:
        A dict containing the grouped logs, with each key of the dict representing the
        version of the log key with equal values, and the value being the equal value.
    """
    api_key = _validate_api_key(api_key)
    headers = {
        "accept": "application/json",
        "Authorization": f"Bearer {api_key}",
    }
    project = _get_and_maybe_create_project(project, api_key=api_key)
    params = {"project": project, "key": key}
    response = requests.get(BASE_URL + "/logs/groups", headers=headers, params=params)
    response.raise_for_status()
    return response.json()


def group_logs(key: str, project: Optional[str] = None, api_key: Optional[str] = None):
    """
    Groups logs based on equality '==' of the values for the specified key, returning a
    dict with group indices as the keys and the list of logs as the values. If the keys
    are not versioned, then the indices are simply incrementing integers.

    Args:
        key: Name of the log entry to do equality matching for.

        project: Name of the project to get logs from.

        api_key: If specified, unify API key to be used. Defaults to the value in the
        `UNIFY_KEY` environment variable.

    Returns:
        A dict containing the grouped logs, with each key of the dict representing the
        version of the log key with equal values, and the value being a list of logs.
    """
    api_key = _validate_api_key(api_key)
    project = _get_and_maybe_create_project(project, api_key=api_key)
    return {
        k: get_logs(
            project,
            "{} == {}".format(key, '"' + v + '"' if isinstance(v, str) else v),
            api_key=api_key,
        )
        for k, v in get_groups(key, project, api_key=api_key).items()
    }


def get_logs_metric(
    metric: str,
    key: str,
    filter: Optional[str] = None,
    project: Optional[str] = None,
    api_key: Optional[str] = None,
) -> Union[float, int, bool]:
    """
    Retrieve a set of log metrics across a project, after applying the filtering.

    Args:
        metric: The reduction metric to compute for the specified key. Supported are:
        sum, mean, var, std, min, max, median, mode.

        key: The key to compute the reduction statistic for.

        filter: The filtering to apply to the various log values, expressed as a string,
        for example:
        "(temperature > 0.5 and (len(system_msg) < 100 or 'no' in usr_response))"

        project: The id of the project to retrieve the logs for.

        api_key: If specified, unify API key to be used. Defaults to the value in the
        `UNIFY_KEY` environment variable.

    Returns:
        The full set of reduced log metrics for the project, after optionally applying
        the optional filtering.
    """
    api_key = _validate_api_key(api_key)
    headers = {
        "accept": "application/json",
        "Authorization": f"Bearer {api_key}",
    }
    project = _get_and_maybe_create_project(project, api_key=api_key)
    params = {"project": project, "filter_expr": filter}
    response = requests.get(
        BASE_URL + f"/logs/metric/{metric}/{key}",
        headers=headers,
        params=params,
    )
    response.raise_for_status()
    return response.json()


def group_logs_by_config(logs: List[unify.Log]) -> Dict:
    configs = list(dict.fromkeys([json.dumps(lg.config.parameters) for lg in logs]))
    ret_dict = dict()
    for conf in configs:
        ret_dict[conf] = [lg for lg in logs if json.dumps(lg.config.parameters) == conf]
    return ret_dict



from typing import List, Optional

from openai.types.chat import ChatCompletion as _ChatCompletion
from openai.types.chat.chat_completion import (
    ChatCompletionMessage,
)
from openai.types.chat.chat_completion import Choice
from openai.types.completion_usage import CompletionUsage


class ChatCompletion(_ChatCompletion):
    # only override pydantic types  which require FormattedBaseModel
    choices: List[Choice]
    usage: Optional[CompletionUsage] = None

    def __init__(self, assistant_message: Optional[str] = None, **kwargs):
        """
        Create ChatCompletion instance.

        Args:
            assistant_message: The assistant message, optional.

        Returns:
            The pydantic ChatCompletion instance.
        """
        if assistant_message:
            kwargs = {
                **dict(
                    id="",
                    choices=[
                        Choice(
                            finish_reason="stop",
                            index=0,
                            message=ChatCompletionMessage(
                                role="assistant",
                                content=assistant_message,
                            ),
                        ),
                    ],
                    created=0,
                    model="",
                    object="chat.completion",
                    **kwargs,
                ),
                **kwargs,
            }
        super().__init__(**kwargs)



import json
import os
import threading
from typing import Any, Dict, List, Optional, Union

import unify
import requests
from pydantic import BaseModel, ValidationError

PROJECT_LOCK = threading.Lock()


def _res_to_list(response: requests.Response) -> Union[List, Dict]:
    return json.loads(response.text)


def _validate_api_key(api_key: Optional[str]) -> str:
    if api_key is None:
        api_key = os.environ.get("UNIFY_KEY")
    if api_key is None:
        raise KeyError(
            "UNIFY_KEY is missing. Please make sure it is set correctly!",
        )
    return api_key


def _default(value: Any, default_value: Any) -> Any:
    return value if value is not None else default_value


def _dict_aligns_with_pydantic(dict_in: Dict, pydantic_cls: type(BaseModel)) -> bool:
    try:
        pydantic_cls.model_validate(dict_in)
        return True
    except ValidationError:
        return False


def _get_and_maybe_create_project(
    project: Optional[str] = None,
    required: bool = True,
    api_key: Optional[str] = None,
) -> Optional[str]:
    api_key = _validate_api_key(api_key)
    if project is None:
        project = unify.active_project
        if project is None:
            if required:
                raise Exception(
                    "No project specified in the arguments, and no globally set "
                    "project either, with required=True was passed. A project must be "
                    "passed in the argument, or set globally via "
                    "unify.activate('project_name')",
                )
            return None
    PROJECT_LOCK.acquire()
    if project not in unify.list_projects(api_key):
        unify.create_project(project, api_key=api_key)
    PROJECT_LOCK.release()
    return project



import inspect
import json
import time
import traceback
import importlib
import os.path
from typing import Optional

import unify
from .system_messages import (
    CODING_SYS_MESSAGE_BASE,
    INIT_CODING_SYS_MESSAGE,
    UPDATING_CODING_SYS_MESSAGE,
    DOCSTRING_SYS_MESSAGE_HEAD,
    DOCSTRING_SYS_MESSAGE_FIRST_CONTEXT,
    DOCSTRING_SYS_MESSAGE_EXTRA_CONTEXT,
    DOCSTRING_SYS_MESSAGE_TAIL,
)

MODEL = "gpt-4o@openai"


IMPLEMENTATIONS = dict()
MODULE_PATH = "implementations.py"

INTERACTIVE = True

FN_TREE = dict()
FN_POINTER = list()


def set_interactive():
    global INTERACTIVE
    INTERACTIVE = True


def set_non_interactive():
    global INTERACTIVE
    INTERACTIVE = False


def interactive_mode():
    global INTERACTIVE
    return INTERACTIVE


class Interactive:

    def __enter__(self):
        set_interactive()

    def __exit__(self, exc_type, exc_val, exc_tb):
        set_non_interactive()


def _formatted(content: str) -> str:
    too_long = True
    while too_long:
        old_content = content
        content = content.replace("\n\n\n\n", "\n\n\n")
        too_long = content != old_content
    return content if (content and content[-1] == "\n") else content + "\n"


def _streamed_print(text: str, *args) -> None:
    for line in text.splitlines():
        for word in line.split(" "):
            print(word + " ", *args, end="")
            time.sleep(0.02)
        print("\n", *args, end="")


def implement(fn: callable, module_path: Optional[str] = None):

    global MODULE_PATH
    if module_path is None:
        module_path = MODULE_PATH
        full_module_path = os.path.join(os.getcwd(), MODULE_PATH)

    def _load_module(module_name: str):
        if not os.path.exists(full_module_path):
            with open(full_module_path, "w+") as file:
                file.write("import unify\n\n\n")
            return
        while True:
            try:
                return importlib.reload(
                    importlib.import_module(module_name),
                )
            except Exception as e:
                _streamed_print(
                    "\nOops, seems like there was an error loading " "our new module.",
                    e,
                )
                _streamed_print(
                    "Don't worry about any undefined imaginary functions which may "
                    "have red underlines shown in your IDE etc., we just need to fix "
                    "the specific issue mentioned above and then "
                    "move to the next step.\n"
                    f"Open the file `{full_module_path}` and fix the issue "
                    "mentioned above, "
                    "(don't forget to ctrl-S to save in your IDE/editor!), "
                    "then press enter once you're happy 👌\n",
                )
                input()

    global IMPLEMENTATIONS
    module = _load_module(module_path.rstrip(".py"))
    for name, obj in inspect.getmembers(module):
        if callable(obj):
            IMPLEMENTATIONS[obj.__name__] = obj

    def _populate_dev_system_message(template: str) -> str:
        return (
            template.replace(
                "{name}",
                name,
            )
            .replace(
                "{docstring}",
                docstring.lstrip("\n").rstrip(" ").rstrip("\n"),
            )
            .replace(
                "{signature}",
                signature,
            )
        )

    def _populate_docstring_context_system_message(
        template: str,
        child_name: str,
        parent_name: str,
        parent_implementation: str,
        calling_line: str,
    ) -> str:
        return (
            template.replace(
                "{child_name}",
                child_name,
            )
            .replace(
                "{parent_name}",
                parent_name,
            )
            .replace(
                "{parent_implementation}",
                parent_implementation,
            )
            .replace(
                "{calling_line}",
                calling_line,
            )
        )

    name = fn.__name__ if fn.__name__ is not None else ""
    docstring = fn.__doc__ if fn.__doc__ is not None else ""
    signature = str(inspect.signature(fn))
    client = unify.Unify(MODEL, cache=True)
    docstring_client = unify.Unify(MODEL, cache=True)
    first_line = f"def {name}{signature}:"
    init_system_message = _populate_dev_system_message(INIT_CODING_SYS_MESSAGE)
    update_system_message = _populate_dev_system_message(UPDATING_CODING_SYS_MESSAGE)
    system_message_base = _populate_dev_system_message(CODING_SYS_MESSAGE_BASE)
    init_system_message += system_message_base
    update_system_message += system_message_base
    client.set_system_message(init_system_message)
    client.set_stateful(True)

    def _get_imports(response):
        top_half = response.split(first_line)[0].rstrip("\n")
        lines = top_half.split("\n")
        imports = list()

        for line in reversed(lines):
            if not line.startswith("import ") and not line.startswith("from "):
                break
            imports.append(line)
        return "\n".join(reversed(imports))

    def _parse_src_code(fn_name, context=None):
        if context is None:
            with open(full_module_path, "r") as file:
                context = file.read()
        lines = context.split("\n")
        starting_line = [f"def {fn_name}(" in ln for ln in lines].index(True)
        valid_lines = lines[starting_line : starting_line + 1]
        indent = len(lines[starting_line + 1]) - len(
            lines[starting_line + 1].lstrip(" "),
        )
        for line in lines[starting_line + 1 :]:
            line_len = len(line)
            if (line_len > indent and line[0:indent] != " " * indent) or (
                line_len <= indent and line.strip(" ").strip("\n") != ""
            ):
                break
            valid_lines.append(line)
        return "\n".join(valid_lines)

    def _remove_docstring(src_code: str) -> str:
        lines = src_code.splitlines()
        removing = False
        new_lines = list()
        for ln in lines:
            if ln.lstrip(" ") == '"""':
                removing = not removing
                continue
            if removing:
                continue
            new_lines.append(ln)
        return "\n".join(new_lines)

    def _inject_docstring(src_code: str, dcstr: str) -> str:
        src_code = _remove_docstring(src_code)
        lines = src_code.splitlines()
        def_idx = ["def " in ln for ln in lines].index(True)
        dcstr_lines = dcstr.splitlines()
        indent = len(dcstr_lines[1]) - len(dcstr_lines[1].lstrip(" "))
        dcstr = "\n".join([" " * (4 - indent) + ln for ln in dcstr_lines])
        return "\n".join(
            lines[0 : def_idx + 1] + [f'    """{dcstr}"""'] + lines[def_idx + 1 :],
        )

    def _generate_code():
        response = client.generate(
            f"please implement the function `{name}` by strictly following the "
            "instructions in the original system message, and also carefully following "
            "any instructions in the full chat history, if any is present:",
        )
        assert (
            first_line in response,
            "Model failed to follow the formatting instructions.",
        )
        fn_names = [
            ln.split("def ")[-1].split("(")[0]
            for ln in response.splitlines()
            if len(ln) > 7 and ln[0:4] == "def "
        ]
        source_codes = {
            fn_name: (
                _inject_docstring(_parse_src_code(fn_name, response), docstring)
                if fn_name == name
                else _parse_src_code(fn_name, response)
            )
            for fn_name in fn_names
        }
        return (
            _get_imports(response),
            source_codes,
            response,
        )

    def _generate_function_spec(fn_name):
        response = docstring_client.generate(
            f"please implement the function definition for `{fn_name}` as described in "
            "the format described in the system message.",
        )
        return _parse_src_code(fn_name, response), response

    def _load_function(fn_name: str):
        while True:
            try:
                return getattr(
                    _load_module(module_path.rstrip(".py")),
                    fn_name,
                )
            except Exception as e:
                _streamed_print(
                    "Hmmm, we loaded module without any errors, "
                    "but there was an error trying to load the function",
                    e,
                )
                _streamed_print(
                    f"Open file `{full_module_path}` and fix the issue mentioned "
                    "above, then press enter once you're done 👌\n",
                )
                input()

    def _write_imports_to_file(imports=""):
        with open(full_module_path, "r") as file:
            content = file.read()
        new_content = imports + content
        with open(full_module_path, "w") as file:
            file.write(_formatted(new_content))

    def _write_functions_to_file(implementations):
        with open(full_module_path, "r") as file:
            content = file.read()
        new_content = content
        for fn_name, implementation in implementations.items():
            if f"def {fn_name}" not in content:
                new_content = new_content.rstrip("\n") + "\n\n\n"
                new_content = new_content + implementation
            else:
                src_code = _parse_src_code(fn_name)
                new_content = new_content.replace(src_code, implementation)
        with open(full_module_path, "w") as file:
            file.write(_formatted(new_content))

    def _remove_unify_decorator_if_present(fn_name: str):
        with open(full_module_path, "r") as file:
            content = file.read()
        lines = content.splitlines()
        fn_line_no = [f"def {fn_name}" in ln for ln in lines].index(True)
        has_decorator = lines[fn_line_no - 1] == "@unify.implement"
        if not has_decorator:
            return
        with open(full_module_path, "w") as file:
            file.write(
                _formatted("\n".join(lines[0 : fn_line_no - 1] + lines[fn_line_no:])),
            )

    def _step_loop(this_client, assistant_msg="") -> str:
        if assistant_msg:
            _streamed_print(assistant_msg)
        assistant_questions = (
            "\nIs there anything you would like me to change? "
            "If so, then please respond in the following format:\n"
            '"Yes: {your explanation}"\n\n'
            "If you would like to make updates yourself, then you can directly "
            f"modify the source code in `{full_module_path}`.\n"
            "Simply respond in the following format once "
            "you've made the changes, and then I can take another look:\n"
            '"Reload: {your explanation}"\n\n'
            "Finally, if you're happy with these edits "
            "and would like to move onto the next function, "
            "then just press enter!\n"
        )
        _streamed_print(assistant_questions)
        response = input().strip("'").strip('"')
        if response == "":
            return ""
        elif response[0:6].lower() == "reload":
            loaded_implementation = _parse_src_code(name)
            this_client.append_messages(
                [
                    {
                        "role": "user",
                        "content": response + ". Hers is the new implementation:"
                        f"\n{loaded_implementation}",
                    },
                ],
            )
            return "reload"
        elif response[0:3].lower() != "yes":
            _streamed_print(
                "Please either press enter to move onto the next step, "
                "or respond in one of the two following formats:\n"
                "Yes: {your explanation}\n"
                "reload: {your explanation}",
            )
            return "invalid"
        this_client.append_messages(
            [
                {"role": "assistant", "content": assistant_questions},
                {"role": "user", "content": response},
            ],
        )
        return "yes"

    def _get_fn_tree_leaf():
        global FN_POINTER
        global FN_TREE
        leaf = FN_TREE
        for key in FN_POINTER:
            if key not in leaf:
                leaf[key] = dict()
            leaf = leaf[key]
        return leaf

    def _add_child_functions_to_state(implementation: str):
        segments = [s.split("(")[0] for s in implementation.split(" ") if "(" in s]
        segments = [
            s
            for s in segments
            if s.replace("_", "").isalnum() and s == s.lower() and s != name
        ]
        leaf = _get_fn_tree_leaf()
        for s in segments:
            leaf[s] = dict()
        if segments:
            return segments[0]

    def _add_args_to_system_msg(*args, **kwargs):
        client.set_system_message(
            client.system_message.replace(
                "{args}",
                str(args)[1:-1],
            ).replace(
                "{kwargs}",
                str(kwargs)[1:-1].replace("'", "").replace(": ", "="),
            ),
        )

    def _get_fn(*args, **kwargs):
        global IMPLEMENTATIONS
        if name in IMPLEMENTATIONS:
            _remove_unify_decorator_if_present(name)
            _streamed_print(f"\n`{name}` is already implemented, stepping inside.\n")
            return IMPLEMENTATIONS[name]
        _streamed_print(
            f"We'll now work together to implement the function `{name}`.\n\n",
        )
        _add_args_to_system_msg(*args, **kwargs)
        imports, implementations, llm_response = _generate_code()
        client.set_system_message(update_system_message)
        _add_args_to_system_msg(*args, **kwargs)
        _write_imports_to_file(imports)
        _write_functions_to_file(implementations)
        if interactive_mode():
            should_continue = True
            assistant_msg = llm_response
            while should_continue:
                mode = _step_loop(client, assistant_msg)
                if mode == "reload":
                    assistant_msg = (
                        "Here is the latest implementation (following any changes "
                        "you may have made):"
                        f"\n\n{_parse_src_code(name)}"
                    )
                should_continue, should_update = {
                    "yes": (True, True),
                    "": (False, False),
                    "reload": (True, False),
                    "invalid": (True, False),
                }[mode]
                if not should_update:
                    continue
                imports, implementations, llm_response = _generate_code()
                assistant_msg = llm_response
                _write_imports_to_file(imports)
                _write_functions_to_file(
                    implementations,
                )
        _remove_unify_decorator_if_present(name)
        fn_implemented = _load_function(name)
        first_child = _add_child_functions_to_state(implementations[name])
        message = (
            "Great! Now that we have an "
            f"implementation in place for `{name}`, "
            "let's decide what to work on next!\n"
            "The total set of functions being worked on within the uppermost "
            "`unify.implement` context are as follows:"
            "\n\n"
            f"{json.dumps(FN_TREE, indent=4).replace('{},', '⏳').replace('{}', '⏳').replace('True,', '✅').replace('True', '✅')}"
            "\n\n"
        )
        if first_child:
            message += f"We will work on implementing `{first_child}` next.\n"
        _streamed_print(message)
        IMPLEMENTATIONS[name] = fn_implemented
        _write_imports_to_file(imports)
        _write_functions_to_file(implementations)
        return fn_implemented

    def _add_new_function_spec_w_implement_decorator(name_error, tracebk):
        system_message = DOCSTRING_SYS_MESSAGE_HEAD.replace("{name}", name_error.name)
        context = list()
        first_context = True
        parent_name = None
        for line in reversed(tracebk.split("\n")):
            if not line or line[0] != " ":
                continue
            context.append(line)
            if line[0:8] == '  File "':
                if first_context:
                    parent_name = context[1].split(",")[-1].split(" ")[-1]
                    system_message += _populate_docstring_context_system_message(
                        DOCSTRING_SYS_MESSAGE_FIRST_CONTEXT,
                        child_name=name_error.name,
                        parent_name=parent_name,
                        parent_implementation=_parse_src_code(parent_name),
                        calling_line=context[0].lstrip(" "),
                    )
                    first_context = False
                else:
                    child_name = parent_name
                    parent_name = context[1].split(",")[-1].split(" ")[-1]
                    system_message += _populate_docstring_context_system_message(
                        DOCSTRING_SYS_MESSAGE_EXTRA_CONTEXT,
                        child_name=child_name,
                        parent_name=parent_name,
                        parent_implementation=_parse_src_code(parent_name),
                        calling_line=context[0].lstrip(" "),
                    )
                context.clear()
                if parent_name == name:
                    break

        system_message += DOCSTRING_SYS_MESSAGE_TAIL.replace("{name}", name_error.name)
        docstring_client.set_system_message(system_message)

    def _get_fn_spec(name_error, tracebk):
        _add_new_function_spec_w_implement_decorator(name_error, tracebk)
        function_spec, llm_response = _generate_function_spec(name_error.name)
        function_spec = "\n\n@unify.implement\n" + function_spec
        _write_functions_to_file({name_error.name: function_spec})
        if interactive_mode():
            should_continue = True
            assistant_msg = llm_response
            while should_continue:
                mode = _step_loop(docstring_client, assistant_msg)
                if mode == "reload":
                    assistant_msg = (
                        "Here is the latest implementation (following any changes "
                        "you may have made):"
                        f"\n\n{_parse_src_code(name)}"
                    )
                should_continue, should_update = {
                    "yes": (True, True),
                    "": (False, False),
                    "reload": (True, False),
                    "invalid": (True, False),
                }[mode]
                if not should_update:
                    continue
                function_spec, llm_response = _generate_function_spec(
                    name_error.name,
                )
                assistant_msg = llm_response
                _write_functions_to_file({name_error.name: function_spec})

        _streamed_print(
            "\nGreat, I'm glad we're both happy with the specification for "
            f"`{name_error.name}`!\n",
        )

    def _execute_with_implement(func: callable, *args, **kwargs):
        try:
            return func(*args, **kwargs)
        except NameError as ne:
            _streamed_print(
                f"\nOkay, so it seems like `{ne.name}` is not yet implemented, "
                "let's define the function specification together first, "
                "before moving onto the implementation in the next step.\n"
                "I'll make a first attempt based on the call stack, and then you "
                "can give me feedback and we can iterate together 🔁\n",
            )
            _get_fn_spec(ne, traceback.format_exc())
        return _load_function(func.__name__)(*args, **kwargs)

    def implemented(*args, **kwargs):
        global FN_POINTER
        FN_POINTER.append(name)
        ret = _execute_with_implement(_get_fn(*args, **kwargs), *args, **kwargs)
        FN_POINTER.pop()
        _get_fn_tree_leaf()[name] = True
        return ret

    return implemented



import json

import unify
from typing import Optional, List, Dict, Any

from .system_messages import SUGGEST_SYS_MESSAGE


def _print_table_from_dicts(dcts, col_list=None) -> str:
    if not col_list:
        col_list = list(dcts[0].keys() if dcts else [])
    my_list = [col_list]
    for item in dcts:
        my_list.append(
            [str(item[col] if item[col] is not None else "") for col in col_list],
        )
    col_size = [max(map(len, col)) for col in zip(*my_list)]
    format_str = " | ".join(["{{:<{}}}".format(i) for i in col_size])
    my_list.insert(1, ["-" * i for i in col_size])
    ret = list()
    for item in my_list:
        ret.append(format_str.format(*item))
    return "\n".join(ret)


def _format_evals(evals: Dict[str, List[Dict[str, Any]]]) -> str:
    ret = list()
    for config_str, logs in evals.items():
        ret.append(
            "Configuration:\n" "--------------\n",
        )
        config = json.loads(config_str)
        for param_name, param_value in config.items():
            ret.append(param_name + ":\n")
            ret.append(param_value)
        ret.append(
            "\n" "Logs:\n" "-----\n",
        )
        ret.append(
            json.dumps(logs, indent=4),
        )
    return "\n".join(ret)


def suggest(
    metric: str,
    mode: str = "maximize",
    logs: Optional[List[unify.Log]] = None,
):
    """
    Make a suggestion for how to improve the performance of the metric specified by
    either the maximize or minimize

    Args:
        metric: The metric name that we're looking to optimize.

        mode: The optimization mode, either maximize or minimize for the metric.

        logs: The logs to parse and use for the suggestion.

    Returns:
          The suggested parameter name to update, and the suggested new value.
    """
    assert mode in ("maximize", "minimize"), "Mode must be 'maximize' or 'minimize'."

    client = unify.Unify("gpt-4o@openai", cache=True)

    if not logs:
        logs = unify.get_logs()

    # ToDo: remove hard-coding once REST API is complete
    config = unify.Parameters(
        evaluator_code="'def evaluate_response(question: str, response: str) -> float:\n    correct_answer = eval(question)\n    try:\n        response_int = int(\n            "
        ".join([c for c in response.split("
        ")[-1] if c.isdigit()]),\n        )\n        return float(correct_answer == response_int)\n    except ValueError:\n        return 0.\n'",
    )
    for lg in logs:
        if "evaluator_code" in lg.entries:
            del lg.entries["evaluator_code"]
        lg._parameters = config
    # End ToDo

    evals = unify.group_logs_by_config(logs)
    # ToDo: fix this hack once ordering is correctly preserved
    evals = {
        k: [
            {
                "question": lg.entries["question"],
                "response": lg.entries["response"],
                "score": lg.entries["score"],
            }
            for lg in logs
        ]
        for k, logs in evals.items()
    }
    # End ToDo
    evals_str = _format_evals(evals)

    system_message = (
        SUGGEST_SYS_MESSAGE.replace(
            "{configs}",
            str(list(config.parameters.keys()))[1:-1],
        )
        .replace(
            "{evals}",
            evals_str,
        )
        .replace(
            "{relation}",
            mode,
        )
        .replace(
            "{low|high}",
            "low" if mode == "maximize" else "high",
        )
        .replace(
            "{metric}",
            metric,
        )
    )
    response = client.generate(
        "please follow the instructions of the system message",
        system_message=system_message,
    )
    breakpoint()
    return response



from __future__ import annotations
import time
import uuid
import datetime

from ..utils.helpers import _validate_api_key
from .utils.logging import _handle_special_types
from .utils.logging import *


# Logs #
# -----#


class Log:

    def __init__(
        self,
        id: int,
        timestamp: Optional[datetime] = None,
        api_key: Optional[str] = None,
        parameters: Dict[str, Any] = None,
        **kwargs,
    ):
        self._id = id
        self._timestamp = timestamp
        self._entries = kwargs
        self._parameters = parameters
        self._api_key = _validate_api_key(api_key)

    # Properties

    @property
    def id(self) -> int:
        return self._id

    @property
    def timestamp(self) -> Optional[datetime]:
        return self._timestamp

    @property
    def entries(self) -> Dict[str, Any]:
        return self._entries

    @property
    def parameters(self) -> Dict[str, Any]:
        return self._parameters

    # Dunders

    def __eq__(self, other: Union[dict, Log]) -> bool:
        if isinstance(other, dict):
            other = Log(other["id"], **other["entries"])
        return self._id == other._id

    def __len__(self):
        return len(self._entries)

    def __repr__(self) -> str:
        return f"Log(id={self._id})"

    # Public

    def download(self):
        self._entries = get_log_by_id(self._id, self._api_key)._entries

    def add_entries(self, **kwargs) -> None:
        add_log_entries(self._id, api_key=self._api_key, **kwargs)
        self._entries = {**self._entries, **kwargs}

    def replace_entries(self, **kwargs) -> None:
        replace_log_entries(self._id, self._api_key, **kwargs)
        self._entries = {**self._entries, **kwargs}

    def update_entries(self, fn, **kwargs) -> None:
        update_log_entries(fn, self._id, self._api_key, **kwargs)
        for k, v in kwargs.items():
            self._entries[k] = fn(self._entries[k], v)

    def rename_entries(self, **kwargs) -> None:
        rename_log_entries(self._id, self._api_key, **kwargs)
        for old_name, new_name in kwargs.items():
            self._entries[new_name] = self._entries[old_name]
            del self._entries[old_name]

    def delete_entries(
        self,
        keys_to_delete: List[str],
    ) -> None:
        for key in keys_to_delete:
            delete_log_entry(key, self._id, self._api_key)
            del self._entries[key]

    def delete(self) -> None:
        delete_log(self._id, self._api_key)

    def to_json(self):
        return {
            "id": self._id,
            "timestamp": self._timestamp,
            "entries": self._entries,
            "parameters": self._parameters,
            "api_key": self._api_key,
        }

    @staticmethod
    def from_json(state):
        entries = state["entries"]
        del state["entries"]
        state = {**state, **entries}
        return Log(**state)


class Context:
    def __init__(self, **kwargs):
        self.kwargs = _handle_special_types(kwargs)

    def __enter__(self):
        self.token = current_global_active_log_kwargs.set(
            {**current_global_active_log_kwargs.get(), **self.kwargs},
        )
        self.nest_level_token = current_context_nest_level.set(
            current_context_nest_level.get() + 1,
        )

    def __exit__(self, *args, **kwargs):
        # print("Before clearing", current_global_active_log_kwargs.get())
        current_global_active_log_kwargs.reset(self.token)
        current_context_nest_level.reset(self.nest_level_token)
        if current_context_nest_level.get() == 0:
            current_logged_logs.set({})


def span(io=True):
    def wrapper(fn):
        def wrapped(*args, **kwargs):
            t1 = time.perf_counter()
            if not current_span.get():
                running_time.set(t1)
            inputs = None
            if io:
                signature = inspect.signature(fn)
                bound_args = signature.bind(*args, **kwargs)
                bound_args.apply_defaults()
                inputs = bound_args.arguments
            new_span = {
                "id": str(uuid.uuid4()),
                "parent_span_id": (
                    None if not current_span.get() else current_span.get()["id"]
                ),
                "span_name": fn.__name__,
                "exec_time": None,
                "offset": round(
                    0.0 if not current_span.get() else t1 - running_time.get(),
                    2,
                ),
                "inputs": inputs,
                "outputs": None,
                "errors": None,
                "child_spans": [],
            }
            token = current_span.set(new_span)
            result = None
            try:
                result = fn(*args, **kwargs)
                return result
            except Exception as e:
                new_span["errors"] = str(e)
                raise e
            finally:
                t2 = time.perf_counter()
                exec_time = t2 - t1
                current_span.get()["exec_time"] = round(exec_time, 2)
                current_span.get()["outputs"] = (
                    None if result is None or not io else result
                )
                if token.old_value is token.MISSING:
                    unify.log(trace=current_span.get(), skip_duplicates=False)
                    current_span.reset(token)
                else:
                    current_span.reset(token)
                    current_span.get()["child_spans"].append(new_span)

        async def async_wrapped(*args, **kwargs):
            t1 = time.perf_counter()
            if not current_span.get():
                running_time.set(t1)
            inputs = None
            if io:
                signature = inspect.signature(fn)
                bound_args = signature.bind(*args, **kwargs)
                bound_args.apply_defaults()
                inputs = bound_args.arguments
            new_span = {
                "id": str(uuid.uuid4()),
                "parent_span_id": (
                    None if not current_span.get() else current_span.get()["id"]
                ),
                "span_name": fn.__name__,
                "exec_time": None,
                "offset": round(
                    0.0 if not current_span.get() else t1 - running_time.get(),
                    2,
                ),
                "inputs": inputs,
                "outputs": None,
                "errors": None,
                "child_spans": [],
            }
            token = current_span.set(new_span)
            # capture the arguments here
            result = None
            try:
                result = await fn(*args, **kwargs)
                return result
            except Exception as e:
                new_span["errors"] = str(e)
                raise e
            finally:
                t2 = time.perf_counter()
                exec_time = t2 - t1
                current_span.get()["exec_time"] = round(exec_time, 2)
                current_span.get()["outputs"] = (
                    None if result is None or not io else result
                )
                if token.old_value is token.MISSING:
                    unify.log(trace=current_span.get(), skip_duplicates=False)
                    current_span.reset(token)
                else:
                    current_span.reset(token)
                    current_span.get()["child_spans"].append(new_span)

        return wrapped if not inspect.iscoroutinefunction(fn) else async_wrapped

    return wrapper
